# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Search Click packages."""

from __future__ import print_function

__metaclass__ = type
__all__ = [
    'ClickRepository',
    'ClickRepositoryError',
    ]

import json
import platform
import os

import apt_pkg
from gi.repository import Click
import pycurl

from io import BytesIO
from io import StringIO

from six.moves.urllib.parse import quote as url_quote

from click.json_helpers import json_array_to_python


class ClickRepositoryError(Exception):
    pass


class HttpResponse:
    """Http Response with status and data members"""

    def __init__(self, status, data):
        self.status = status
        self.data = data

    def __unicode__(self):
        return self.data


# similar to httplib2.Http().request()
def http_request(url, data=None, headers=None):
    """Send a http request to the given url and return (response, data)

    param url  The url to send the request to
    param data  (Optional) data to POST to the url
    param headers  (Optional) headers to include with the request
    return  A tuple of (response, body)
    rtype  (HttpResponse, bytes)
    """
    def write_header(line):
        header.write(line.decode("utf-8"))
        return len(line)
    def write_body(data):
        body.write(data)
        return len(data)
    def post_data(num):
        return data_stream.read(num)
    header = StringIO()
    body = BytesIO()
    curl = pycurl.Curl()
    curl.setopt(curl.URL, url)
    if os.environ.get("CLICK_DEBUG_HTTP", ""):    
        curl.setopt(curl.VERBOSE, 1)
    curl.setopt(pycurl.WRITEFUNCTION, write_body)
    curl.setopt(pycurl.HEADERFUNCTION, write_header)
    if data is not None:
        data_stream = BytesIO(data)
        curl.setopt(pycurl.READFUNCTION, post_data)
        curl.setopt(pycurl.POSTFIELDSIZE, len(data))
        curl.setopt(pycurl.POST, 1)
    if headers is not None:
        header_list = []
        for k, v in headers.items():
            header_list.append("%s: %s" % (k, v))
        curl.setopt(pycurl.HTTPHEADER, header_list)
    # ssl: no need to set extra options, pycurl has sensible defaults
    curl.perform()
    resp = HttpResponse(curl.getinfo(pycurl.HTTP_CODE), header.getvalue())
    return resp, body.getvalue()


def get_repository():
    """Factory to get a click repository for the current distro"""
    distro = platform.dist()[0]
    if distro  == 'Ubuntu':
        return ClickRepositoryAppsUbuntuCom()
    raise ClickRepositoryError("No repository for distribution '%s'" % distro)


class ClickRepository:

    def search(self, search_string):
        raise NotImplementedError()

    def upgradable(self):
        raise NotImplementedError()


class ClickRepositoryAppsUbuntuCom(ClickRepository):

    SEARCH_URI = "https://search.apps.ubuntu.com/api/v1/search?q=%s"
    APPS_URI = "https://search.apps.ubuntu.com/api/v1/package/%s"

    def search(self, search_term):
        result = []
        resp, raw_content = http_request(
            self.SEARCH_URI % url_quote(search_term))
        content = raw_content.decode("utf-8")
        if resp.status != 200:
            raise ClickRepositoryError(content)
        data = json.loads(content)
        if not "_embedded" in data:
            return result
        for item in data["_embedded"]["clickindex:package"]:
            result.append(item)
        return result

    def details(self, name):
        resp, raw_content = http_request(
            self.APPS_URI % url_quote(name))
        content = raw_content.decode("utf-8")
        if resp.status != 200:
            raise ClickRepositoryError(content)
        return json.loads(content)

    def upgradable(self):
        result = []
        db = Click.DB()
        db.read(db_dir=None)
        registry = Click.User.for_user(db, name=None)
        for current in json_array_to_python(registry.get_manifests()):
            resp, raw_content = http_request(self.APPS_URI % current["name"])
            content = raw_content.decode("utf-8")
            if resp.status != 200:
                continue
            new = json.loads(content)
            if apt_pkg.version_compare(current["version"], new["version"]) < 0:
                result.append((current, new))
        return result
